"""Containers instantiate objects per specs set in configs."""

from __future__ import annotations

from typing import Any, Generic, Iterable, TypeVar, cast

from typing_extensions import override

import dilib.config
import dilib.specs
import dilib.utils

T = TypeVar("T")
TC = TypeVar("TC", bound=dilib.config.Config)


class ConfigProxy(Generic[TC]):
    """Read-only helper to marshal config values.

    See `patterns.md` for a discussion on why this was necessary.

    :meta private:
    """

    def __init__(
        self, container: Container[TC], config: dilib.config.Config
    ) -> None:
        self.container = container
        self.config = config

    def __getattr__(self, key: str) -> Any:
        # noinspection PyProtectedMember
        return self.container._get(self.config, key)

    def __getitem__(self, key: str) -> Any:
        return dilib.utils.nested_getattr(self, key)

    def __contains__(self, key: str) -> bool:
        return key in self.config

    @override
    def __dir__(self) -> Iterable[str]:
        return dir(self.config)


class Container(Generic[TC]):
    """Materializes and caches (if necessary) objects based on given config.

    Config user should use :func:`get_container` to instantiate.
    """

    def __init__(self, config: TC) -> None:
        if isinstance(config, dilib.config.ConfigSpec):
            raise ValueError(
                "Expected Config type, got ConfigSpec. "
                "Please call .get() on the config."
            )

        self._config = config

        # Once we pass a config to a container, we can no longer
        # perturb it (as this would require updating container caches).
        self._config.freeze()

        self._instance_cache: dict[str | int, Any] = {}

    @property
    def config(self) -> TC:
        """More type-safe alternative to attr access."""
        # Cast because `ConfigProxy[TC]` will act like `TC`.
        return cast(TC, ConfigProxy(self, self._config))

    # noinspection PyProtectedMember
    def _process_arg_spec(
        self, config: dilib.config.Config, arg: dilib.specs.Spec[Any]
    ) -> Any:
        if arg.spec_id in config._keys:
            config_key = config._keys[arg.spec_id]
            result = self._get(config, config_key)
        elif isinstance(arg, dilib.specs._Callable):
            # Anonymous prototype or singleton.
            result = self._materialize_callable_spec(config, arg).instantiate()
        elif isinstance(arg, dilib.specs._Object):
            return arg.obj
        else:
            for child_config in config._child_configs.values():
                if arg.spec_id in child_config._keys:
                    return self._process_arg(child_config, arg)

            raise TypeError(f"Unrecognized arg type: {type(arg)}")

        return result

    # noinspection PyProtectedMember
    def _process_arg(self, config: dilib.config.Config, arg: Any) -> Any:
        if isinstance(arg, dilib.specs.Spec):
            return self._process_arg_spec(config, arg)
        elif isinstance(arg, dilib.specs.AttrFuture):
            config_key = config._keys[arg.root_spec_id]
            result = self._get(config, config_key)

            for attr in arg.attrs:
                result = getattr(result, attr)
        elif isinstance(arg, (tuple, list)):
            result = type(arg)(self._process_arg(config, elem) for elem in arg)
        elif isinstance(arg, dict):
            result = {k: self._process_arg(config, v) for k, v in arg.items()}
        else:
            result = arg

        return result

    # noinspection PyProtectedMember
    def _materialize_callable_spec(
        self, config: dilib.config.Config, spec: dilib.specs._Callable[Any]
    ) -> dilib.specs._Callable[Any]:
        """Return Spec copy with materialized args/kwargs."""
        materialized_args = [
            self._process_arg(config, arg) for arg in spec.args
        ]
        materialized_kwargs = {
            key: self._process_arg(config, arg)
            for key, arg in spec.kwargs.items()
        }
        if spec.lazy_kwargs:
            materialized_lazy_kwargs = self._process_arg(
                config, spec.lazy_kwargs
            )
            materialized_kwargs.update(
                {
                    key: self._process_arg(config, arg)
                    for key, arg in materialized_lazy_kwargs.items()
                }
            )

        return spec.copy_with(*materialized_args, **materialized_kwargs)

    # noinspection PyProtectedMember
    def _get(self, config: dilib.config.Config, key: str) -> Any:
        """Get instance represented by key in given config."""
        spec = getattr(config, key)
        if isinstance(spec, dilib.specs._Object):
            return spec.obj
        elif isinstance(spec, dilib.specs._Singleton):
            try:
                return self._instance_cache[spec.spec_id]
            except KeyError:
                pass

            instance = self._materialize_callable_spec(
                config, spec
            ).instantiate()
            self._instance_cache[spec.spec_id] = instance
            return instance
        elif isinstance(spec, dilib.specs._Prototype):
            return self._materialize_callable_spec(config, spec).instantiate()
        elif isinstance(spec, dilib.config.Config):
            return ConfigProxy(self, spec)
        elif isinstance(spec, dilib.specs.AttrFuture):
            key = config._keys[spec.root_spec_id]
            obj = self._get(config, key)

            for idx, attr in enumerate(spec.attrs):
                obj = getattr(obj, attr)
                if idx == len(spec.attrs) - 1:
                    return obj

            raise ValueError(
                f"Failed to resolve attr reference: "
                f"spec_id={spec.spec_id}, attrs={spec.attrs}"
            )
        else:
            raise ValueError(
                f"Unrecognized spec type: {type(spec)} with key={key!r}"
            )

    def get(self, key: str, default: Any | None = None) -> Any:
        """Get materialized object aliased by key, with optional default."""
        if key in dir(self):
            return self[key]

        return default

    def clear(self) -> None:
        """Clear instance cache."""
        self._instance_cache.clear()

    def __getattr__(self, key: str) -> Any:
        return self._get(self._config, key)

    def __getitem__(self, key: str) -> Any:
        return dilib.utils.nested_getattr(self, key)

    def __contains__(self, key: str) -> bool:
        return dilib.utils.nested_contains(self._config, key)

    @override
    def __dir__(self) -> Iterable[str]:
        return dir(self._config)


def get_container(config: TC) -> Container[TC]:
    """Get instance of container object, which instantiates objects per specs.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.GlobalInput(type_=int)
    ...     y = dilib.Singleton(lambda x: x + 1, x=x)

    >>> config = dilib.get_config(FooConfig, x=1)
    >>> container = dilib.get_container(config)
    >>> container.config.y
    2
    """
    return Container(config)



from dilib.config import Config as Config
from dilib.config import get_config as get_config
from dilib.container import ConfigProxy as ConfigProxy
from dilib.container import Container as Container
from dilib.container import get_container as get_container
from dilib.errors import ConfigError as ConfigError
from dilib.errors import FrozenConfigError as FrozenConfigError
from dilib.errors import InputConfigError as InputConfigError
from dilib.errors import NewKeyConfigError as NewKeyConfigError
from dilib.errors import PerturbSpecError as PerturbSpecError
from dilib.errors import SetChildConfigError as SetChildConfigError
from dilib.specs import Forward as Forward
from dilib.specs import GlobalInput as GlobalInput
from dilib.specs import LocalInput as LocalInput
from dilib.specs import Object as Object
from dilib.specs import Prototype as Prototype
from dilib.specs import PrototypeMixin as PrototypeMixin
from dilib.specs import Singleton as Singleton
from dilib.specs import SingletonDict as SingletonDict
from dilib.specs import SingletonList as SingletonList
from dilib.specs import SingletonMixin as SingletonMixin
from dilib.specs import SingletonTuple as SingletonTuple
from dilib.specs import Spec as Spec
from dilib.specs import SpecID as SpecID
from dilib.specs import config_context as config_context



# mypy: disable-error-code="comparison-overlap"
from __future__ import annotations

from typing import Any, TypeVar, cast

import pytest

import dilib
from dilib.tests import test_config

TC = TypeVar("TC", bound=dilib.Config)


def get_container_objs(
    config: TC | type[TC], more_type_safe: bool, **global_inputs: Any
) -> tuple[dilib.Container[TC], TC]:
    config_proxy: Any
    if more_type_safe:
        if not isinstance(config, dilib.Config):
            config = dilib.get_config(config, **global_inputs)
        elif global_inputs:
            raise ValueError("Cannot set config obj and global inputs")

        container = dilib.get_container(config)

        config_proxy = container.config
    else:
        if not isinstance(config, dilib.Config):
            config = config().get(**global_inputs)
        elif global_inputs:
            raise ValueError("Cannot set config obj and global inputs")

        assert isinstance(config, dilib.Config)
        container = dilib.Container(config)

        config_proxy = container

    # Cast because container will act like `TC`.
    return cast(dilib.Container[TC], container), cast(TC, config_proxy)


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_basic(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.BasicConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.x == 1
    assert config_proxy.y == 2
    assert isinstance(config_proxy.foo, test_config.SingletonValueWrapper)
    assert config_proxy.foo.value == 1
    assert isinstance(config_proxy.bar, test_config.PrototypeValueWrapper)
    assert config_proxy.bar.value == 2

    assert config_proxy.foo is config_proxy.foo  # `foo` is a `Singleton`
    assert config_proxy.bar is not config_proxy.bar  # `foo` is a `Prototype`


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_basic(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.BasicConfig, more_type_safe=more_type_safe
    )

    config.x = 2

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)
    assert config_proxy.y == 3
    assert config_proxy.foo.value == 2
    assert config_proxy.bar.value == 3


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_get_nested(more_type_safe: bool) -> None:
    container, config_proxy = get_container_objs(
        test_config.GrandParentConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.parent_config0.basic_config.x == 1
    assert container["parent_config0.basic_config.x"] == 1


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_dir(more_type_safe: bool) -> None:
    container, config_proxy = get_container_objs(
        test_config.GrandParentConfig, more_type_safe=more_type_safe
    )

    assert (
        dir(container)
        == dir(config_proxy)
        == [
            "foobar",
            "parent_config0",
            "parent_config1",
            "some_str0",
        ]
    )
    assert dir(config_proxy.parent_config0) == ["basic_config", "baz0"]
    assert dir(container["parent_config0"]) == ["basic_config", "baz0"]


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.GrandParentConfig, more_type_safe=more_type_safe
    )

    config.parent_config0.basic_config.x = 10

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.parent_config0.basic_config.x == 10
    assert config_proxy.parent_config0.basic_config.foo.value == 10
    assert config_proxy.parent_config0.basic_config.bar.value == 11
    assert (
        config_proxy.parent_config0.basic_config.foo
        is config_proxy.parent_config1.basic_config.foo
    )
    assert config_proxy.parent_config0.baz0.value == 10
    assert config_proxy.parent_config1.baz1.value == 10
    assert config_proxy.foobar.value == 10


class MoreComplexPerturbConfig0(dilib.Config):
    x: int = dilib.GlobalInput(type_=int)
    y: int = dilib.Object(2)

    foo = dilib.Singleton(test_config.ValueWrapper, 100)


class MoreComplexPerturbConfig1(dilib.Config):
    other_config = MoreComplexPerturbConfig0()

    value_obj = dilib.Singleton(test_config.ValueWrapper, other_config.x)


class MoreComplexPerturbConfig2(dilib.Config):
    other_config = MoreComplexPerturbConfig1()

    value_obj = dilib.Singleton(
        test_config.ValueWrapper, other_config.other_config.x
    )


class Doubler:
    def __init__(self, value: int) -> None:
        self.value = value * 2


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_more_complex_input(more_type_safe: bool) -> None:
    config = test_config.get_config(
        MoreComplexPerturbConfig1, more_type_safe=more_type_safe, x=100
    )

    config.value_obj = dilib.Singleton(  # type: ignore
        Doubler,
        config.other_config.x,
    )

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.value_obj.value == 200


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_more_complex_object0(more_type_safe: bool) -> None:
    config = test_config.get_config(
        MoreComplexPerturbConfig1, more_type_safe=more_type_safe, x=100
    )

    config.value_obj = dilib.Singleton(  # type: ignore
        Doubler,
        config.other_config.y,
    )

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.value_obj.value == 4


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_more_complex_object1(more_type_safe: bool) -> None:
    config = test_config.get_config(
        MoreComplexPerturbConfig2, more_type_safe=more_type_safe, x=100
    )

    config.value_obj = dilib.Singleton(  # type: ignore
        Doubler, config.other_config.other_config.y
    )

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.value_obj.value == 4


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_nested_keyerror(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.ErrorGrandParentConfig, more_type_safe=more_type_safe
    )

    with pytest.raises(KeyError):
        try:
            _ = config_proxy.foobar
        except KeyError as exc:
            assert str(exc) == (
                "\"<class 'dilib.tests.test_config.ParentConfig0'>: "
                "'non_existent_field'\""
            )
            raise


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_input_config(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.InputConfig1, more_type_safe=more_type_safe, name="hi"
    )

    config.input_config0.x = 100

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)
    assert config_proxy.input_config0.name == "hi"
    assert config_proxy.input_config0.context == "default"
    assert config_proxy.input_config0.x == 100
    assert config_proxy.y == 101


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_collection_config(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.CollectionConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.x == 1
    assert config_proxy.y == 2
    assert config_proxy.foo_tuple == (1, 2)
    assert config_proxy.foo_list == [1, 2]
    assert config_proxy.foo_dict_kwargs == {"x": 1, "y": 2}
    assert config_proxy.foo_dict_values0 == {1: 1, 2: 2}
    assert config_proxy.foo_dict_values1 == {"values": 1}
    assert config_proxy.foo_dict_values2 == {"x": 1, "y": 2, "z": 3}

    assert config_proxy.foo_tuple is config_proxy.foo_tuple


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_anonymous(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.AnonymousConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.y.value.value is config_proxy.x
    assert config_proxy.z.value.value is config_proxy.x


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_underscore(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.WrapperConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.value.value is config_proxy._value


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_forward(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        test_config.ForwardConfig, more_type_safe=more_type_safe
    )

    assert (
        config_proxy.foo
        is config_proxy.other_config.parent_config0.basic_config.foo
    )
    assert (
        config_proxy.foo_value.value
        is config_proxy.other_config.parent_config0.basic_config.foo
    )


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_forward(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.ForwardConfig, more_type_safe=more_type_safe
    )

    config.x = 1000

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    # Objs that depend on original x remain unperturbed, but objs
    # that depend on the forward alias are perturbed.
    assert config_proxy.other_config.parent_config0.basic_config.x == 1
    assert config_proxy.other_config.foobar.value == 1
    assert config_proxy.x == 1000
    assert config_proxy.x_value.value == 1000


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_partial_kwargs(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.PartialKwargsConfig, more_type_safe=more_type_safe
    )

    config.x = 10
    config.y = 20

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.x == 10
    assert config_proxy.y == 20
    assert config_proxy.values.x == 10
    assert config_proxy.values.y == 20
    assert config_proxy.values.z == 10


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_partial_kwargs_other(more_type_safe: bool) -> None:
    config = test_config.get_config(
        test_config.PartialKwargsOtherConfig, more_type_safe=more_type_safe
    )

    config.partial_kwargs_config.x = 10

    _, config_proxy = get_container_objs(config, more_type_safe=more_type_safe)

    assert config_proxy.partial_kwargs_config.values.x == 10
    assert config_proxy.partial_kwargs_config.values.y == 2
    assert config_proxy.partial_kwargs_config.values.z == 10

    assert config_proxy.values.x == 10
    assert config_proxy.values.y == 2
    assert config_proxy.values.z == 3


class ObjWithAttr:
    @property
    def test_attr(self) -> int:
        return 1


class ObjAttrConfig(dilib.Config):
    test_obj = dilib.Singleton(ObjWithAttr)
    test_obj_attr = dilib.Forward(test_obj.test_attr)


class NestedConfigObjAttrConfig(dilib.Config):
    cfg = ObjAttrConfig()

    test_obj = dilib.Forward(cfg.test_obj)
    test_obj_attr = dilib.Forward(cfg.test_obj_attr)


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_obj_attr(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        ObjAttrConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.test_obj.test_attr == 1
    assert config_proxy.test_obj_attr == 1


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_nested_config_obj_attr(more_type_safe: bool) -> None:
    _, config_proxy = get_container_objs(
        NestedConfigObjAttrConfig, more_type_safe=more_type_safe
    )

    assert config_proxy.test_obj.test_attr == 1
    assert config_proxy.test_obj_attr == 1
    assert config_proxy.test_obj is config_proxy.cfg.test_obj


def test_typing() -> None:
    config = dilib.get_config(test_config.ParentConfig1)

    # Would trigger mypy (and PyCharm, since we're using get_container) error:
    # container: dilib.Container[
    #     test_config.BasicConfig,
    # ] = dilib.get_container(config)

    container: dilib.Container[test_config.ParentConfig1] = (
        dilib.get_container(config)
    )

    # Would trigger mypy error:
    # x: str = container.config.basic_config.x

    x: int = container.config.basic_config.x
    assert x == 1
    y: int = container.config.basic_config.y
    assert y == 2


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_contains(more_type_safe: bool) -> None:
    container, config_proxy = get_container_objs(
        test_config.GrandParentConfig, more_type_safe=more_type_safe
    )

    for exist_key in ["parent_config1", "parent_config1.basic_config.x"]:
        assert exist_key in container
        assert exist_key in config_proxy
        assert exist_key in container._config

    for missing_key in ["foo_bar", "foo_bar.x", "parent_config1.foo_bar"]:
        assert missing_key not in container
        assert missing_key not in config_proxy
        assert missing_key not in container._config


class PerturbSpecConfig(dilib.Config):
    foo = dilib.Singleton(test_config.ValuesWrapper, 1, 2, z=3)


def test_perturb_spec() -> None:
    config = dilib.get_config(PerturbSpecConfig)
    with pytest.raises(dilib.PerturbSpecError):
        config.foo.x = 100  # type: ignore[misc]



"""Errors."""


class ConfigError(RuntimeError):
    """Base class for all errors."""

    pass


class FrozenConfigError(ConfigError):
    """Cannot perturb config once frozen.

    Note that :func:`dilib.container.get_container`
    automatically freezes given config.
    """

    pass


class InputConfigError(ConfigError):
    """Invalid global or local inputs.

    See error message for more info. E.g., the config user could have
    forgotten to include all required inputs.
    """

    pass


class NewKeyConfigError(ConfigError):
    """Cannot try to set new key on config not described by config author."""

    pass


class SetChildConfigError(ConfigError):
    """Cannot set child configs."""

    pass


class PerturbSpecError(ConfigError):
    """Cannot perturb spec fields (only *config* fields can be perturbed)."""

    pass



"""Configs contain named specs and their dependencies."""

from __future__ import annotations

from typing import Any, Iterable, TypeVar, cast

from typing_extensions import override

import dilib.errors
import dilib.specs
import dilib.utils

T = TypeVar("T")
TC = TypeVar("TC", bound="Config")


class ConfigSpec(dilib.specs.Spec[TC]):
    """Represents nestable bag of types and values.

    :meta private:
    """

    _INTERNAL_FIELDS = dilib.specs.Spec._INTERNAL_FIELDS + [
        "cls",
        "local_inputs",
    ]

    def __init__(self, cls: type[TC], **local_inputs: Any) -> None:
        super().__init__()
        self.cls = cls
        self.local_inputs = local_inputs

    def get(self, **global_inputs: Any) -> Config:
        """Instantiate with given global inputs."""
        config_locator = ConfigLocator(**global_inputs)
        config = config_locator.get(self)

        # noinspection PyProtectedMember
        global_input_keys = config._get_all_global_input_keys()
        extra_global_input_keys = set(global_inputs.keys()) - global_input_keys
        if extra_global_input_keys:
            raise dilib.errors.InputConfigError(
                f"Provided extra global inputs "
                f"not specified in configs: {extra_global_input_keys}"
            )

        return config

    @override
    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is ConfigSpec
            and self.cls is other.cls
            and self.local_inputs == other.local_inputs
        )

    @override
    def __hash__(self) -> int:
        return hash(
            (
                self.__class__.__name__,
                frozenset(self.local_inputs.items()),
            )
        )


class Config:
    """Description of specs and how they depend on each other.

    Config author should subclass this class and describe specs
    like fields of a dataclass (with optional type annotation).

    Config user should use :func:`get_config` to instantiate.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.Object(1)
    ...     y = dilib.Singleton(lambda x: x + 1)
    """

    _INTERNAL_FIELDS = [
        "_config_locator",
        "_keys",
        "_specs",
        "_child_configs",
        "_global_inputs",
        "_loaded",
        "_frozen",
        "_get_all_global_input_keys",
        "_process_input",
        "_load",
        "freeze",
        "_get_spec",
        "_get_child_class",
    ]

    def __new__(
        cls: type[TC], *args: Any, _materialize: bool = False, **kwargs: Any
    ) -> TC:
        if _materialize:
            return super().__new__(cls)  # pyright: ignore[reportArgumentType]
        else:
            if args:
                raise ValueError("args must be empty")

            return ConfigSpec(cls, **kwargs)  # type: ignore[return-value]

    def __init__(
        self,
        config_locator: ConfigLocator | None = None,
        **local_inputs: Any,
    ) -> None:
        if config_locator is None:
            raise ValueError("Use config.get() to get instance of config")
        self._config_locator = config_locator

        # spec id -> spec key
        self._keys: dict[dilib.specs.SpecID, str] = {}
        # key -> spec
        self._specs: dict[str, dilib.specs.Spec[Any]] = {}
        # child config key -> child config
        self._child_configs: dict[str, Config] = {}
        # global input key -> spec id
        self._global_inputs: dict[str, dilib.specs.SpecID] = {}

        self._loaded = False
        self._frozen = False

        self._load(**local_inputs)

    # For mypy
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def _get_all_global_input_keys(
        self,
        all_global_input_keys: dict[str, dilib.specs.SpecID] | None = None,
    ) -> set[str]:
        """Recursively get all global input keys of self and its children."""
        all_global_input_keys = (
            all_global_input_keys if all_global_input_keys is not None else {}
        )

        for key, spec_id in self._global_inputs.items():
            if key in all_global_input_keys:
                if all_global_input_keys[key] != spec_id:
                    raise dilib.errors.InputConfigError(
                        f"Found global input collision: {key!r}"
                    )

            all_global_input_keys[key] = spec_id

        for _, child_config in self._child_configs.items():
            # noinspection PyProtectedMember
            child_config._get_all_global_input_keys(all_global_input_keys)

        return set(all_global_input_keys.keys())

    # noinspection PyProtectedMember
    def _process_input(
        self,
        key: str,
        spec: dilib.specs._Input[Any],
        inputs: dict[str, Any],
        desc: str,
    ) -> dilib.specs._Object[Any]:
        """Convert Input spec to Object spec."""
        try:
            value = inputs[key]
        except KeyError:
            if spec.default != dilib.specs.MISSING:
                value = spec.default
            else:
                raise dilib.errors.InputConfigError(
                    f"{desc} input not set: {key!r}"
                ) from None

        dilib.utils.check_type(value, spec.type_, desc=desc)

        # Preserve old spec id.
        return dilib.specs._Object(value, spec_id=spec.spec_id)

    def _load(self, **local_inputs: Any) -> None:
        """Transfer class variables to instance."""
        for key in self.__class__.__dict__:
            if (
                key.startswith("__")
                or key == "_INTERNAL_FIELDS"
                or key in self._INTERNAL_FIELDS
            ):
                continue

            spec = getattr(self.__class__, key)

            # Skip partial kwargs (no registration needed).
            if isinstance(spec, dict):
                continue

            if not isinstance(spec, dilib.specs.Spec):
                raise ValueError(
                    f"Expected Spec type, got {type(spec)} with {key!r}"
                )

            # Register key.
            self._keys[spec.spec_id] = key

            # Handle inputs.
            # noinspection PyProtectedMember
            if isinstance(spec, dilib.specs._GlobalInput):
                self._global_inputs[key] = spec.spec_id

                spec = self._process_input(
                    key, spec, self._config_locator.global_inputs, "Global"
                )
            # noinspection PyProtectedMember.
            elif isinstance(spec, dilib.specs._LocalInput):
                spec = self._process_input(key, spec, local_inputs, "Local")

            # Handle child configs
            if isinstance(spec, ConfigSpec):
                child_config = self._config_locator.get(spec)
                self._child_configs[key] = child_config
            else:
                self._specs[key] = spec

        self._loaded = True

    def freeze(self) -> None:
        """Prevent any more perturbations to this `Config` instance."""
        self._frozen = True

    def _get_spec(self, key: str) -> dilib.specs.Spec[Any]:
        """More type-safe alternative to get spec than attr access."""
        spec = self[key]
        if not isinstance(spec, dilib.specs.Spec):
            raise TypeError(type(spec))
        return spec

    def _get_child_config(self, key: str) -> Config:
        """More type-safe alternative to get child config than attr access."""
        child_config = self[key]
        if not isinstance(child_config, Config):
            raise TypeError(type(child_config))
        return child_config

    # NB: Have to override `getattribute` instead of `getattr` to
    # prevent initial, class-level values from being used.
    @override
    def __getattribute__(self, key: str) -> Any:
        if (
            key.startswith("__")
            or key == "_INTERNAL_FIELDS"
            or key in self._INTERNAL_FIELDS
        ):
            return super().__getattribute__(key)

        try:
            if key in self._child_configs:
                return self._child_configs[key]
            else:
                return self._specs[key]
        except KeyError:
            raise KeyError(f"{self.__class__}: {key!r}") from None

    def __getitem__(self, key: str) -> Any:
        return dilib.utils.nested_getattr(self, key)

    def __contains__(self, key: str) -> Any:
        if "." in key:
            return dilib.utils.nested_contains(self, key)
        else:
            return key in dir(self)

    @override
    def __setattr__(self, key: str, value: Any) -> None:
        if (
            key.startswith("__")
            or key == "_INTERNAL_FIELDS"
            or key in self._INTERNAL_FIELDS
        ):
            return super().__setattr__(key, value)

        if self._frozen:
            raise dilib.errors.FrozenConfigError(
                f"Cannot perturb frozen config: key={key!r}"
            )

        if key not in self._specs and self._loaded:
            if key in self._child_configs:
                raise dilib.errors.SetChildConfigError(
                    f"Cannot set child config: key={key!r}"
                )
            else:
                raise dilib.errors.NewKeyConfigError(
                    f"Cannot add new keys to a loaded config: key={key!r}"
                )

        old_spec = self._specs[key]

        # Automatically wrap input if user hasn't done so.
        if not isinstance(value, dilib.specs.Spec):
            value = dilib.specs.Object(value)

        self._specs[key] = value

        # Transfer old spec id.
        value.spec_id = old_spec.spec_id

    def __setitem__(self, key: str, value: Any) -> None:
        dilib.utils.nested_setattr(self, key, value)

    @override
    def __dir__(self) -> Iterable[str]:
        return sorted(
            list(self._specs.keys()) + list(self._child_configs.keys())
        )


class ConfigLocator:
    """Service locator to get instances of `Config` objects by type.

    :meta private:
    """

    def __init__(self, **global_inputs: Any) -> None:
        self.global_inputs: dict[str, Any] = global_inputs

        self._config_cache: dict[ConfigSpec[Any], Config] = {}

    def get(self, config_spec: ConfigSpec[Any]) -> Config:
        """Get Config instance by type."""
        try:
            return self._config_cache[config_spec]
        except KeyError:
            pass

        config = dilib.specs.instantiate(
            config_spec.cls,
            self,
            _materialize=True,
            **config_spec.local_inputs,
        )
        self._config_cache[config_spec] = config
        return cast(Config, config)


def get_config(config_cls: type[TC], **global_inputs: Any) -> TC:
    """Get instance of config object (that can optionally be perturbed).

    User is required to pass in any non-defaulted global inputs.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.GlobalInput(type_=int)
    ...     y = dilib.Singleton(lambda x: x + 1)

    >>> config = dilib.get_config(FooConfig, x=1)
    >>> container = dilib.get_container(config)

    See :class:`Config`.
    """
    return cast(TC, config_cls().get(**global_inputs))



# ruff: noqa: UP006
from __future__ import annotations

import abc
import dataclasses
import functools
import itertools
from typing import (
    TYPE_CHECKING,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Set,
    Type,
    TypeVar,
    cast,
    get_type_hints,
    overload,
)
import weakref

from typing_extensions import Self, override

T = TypeVar("T")
R = TypeVar("R")

PRIVATE_CONTAINER_FIELD_NAMES = {
    "_frozen",
    "_instance_cache",
    "_parent_ctrs",
    "_field_keys",
    "_child_ctr_keys",
    "_property_keys",
}


class ContainerError(RuntimeError):
    pass


class ContainerParamsError(ContainerError):
    """Either missing or extra container param."""

    pass


class FrozenContainerError(ContainerError):
    """Once frozen, a container cannot be perturbed.

    Freezing happens when either (1) the user call `Container.freeze()`,
    or (2) whenever any value in the container hierarchy is gotten.
    The latter is to ensure a self-consistency guarantee
    (i.e., every perturb is guaranteed to be reflected in every
    downstream object).
    """

    pass


class NewContainerKeyError(ContainerError):
    """Cannot add a new key not specified by container author."""

    pass


def nested_func(
    ctr: Container, key: str, func: Callable[[Container, str], R]
) -> R:
    key_split = key.split(".")
    for idx, key_part in enumerate(key_split):
        if idx < len(key_split) - 1:
            obj = getattr(ctr, key_part)
            if not isinstance(obj, Container):
                raise TypeError(type(obj))

            ctr = obj
        else:
            return func(ctr, key_part)

    raise RuntimeError("Reached unexpected point")


@dataclasses.dataclass()
class Container:
    """Create and cache (if necessary) objects.

    The author of the child `Container` class describes a universe
    of objects that depend on each other, and the container user can retrieve
    any object in this universe by name.
    """

    _frozen: bool = dataclasses.field(
        default=False, init=False, hash=False, compare=False, repr=False
    )
    _instance_cache: Dict[str, object] = dataclasses.field(
        default_factory=dict, init=False, hash=False, compare=False, repr=False
    )
    _parent_ctrs: ContainerWeakSet = dataclasses.field(
        default_factory=weakref.WeakSet,
        init=False,
        hash=False,
        compare=False,
        repr=False,
    )

    _field_keys: ClassVar[Set[str]]
    _child_ctr_keys: ClassVar[Set[str]]
    _property_keys: ClassVar[Set[str]]

    @property
    def _child_ctrs(self) -> Iterable[Container]:
        for key in self._child_ctr_keys:
            yield getattr(self, key)

    @functools.cached_property
    def _keys(self) -> Set[str]:
        if hasattr(self.__class__, "_property_keys"):
            return self._field_keys.union(self._property_keys)
        else:
            return self._field_keys

    def keys(self) -> Iterable[str]:
        """Available fields and properties (shallow).

        By shallow, we mean we don't recurse down to child containers,
        which are also available keys to the user.
        """
        return self._keys

    def freeze(self) -> None:
        """Prevent any further perturbations."""
        if self._frozen:
            return

        self._frozen = True

        for ctr in itertools.chain(self._parent_ctrs, self._child_ctrs):
            ctr.freeze()

    def _check_not_frozen(self) -> None:
        if self._frozen:
            raise FrozenContainerError(
                "Container is already frozen, "
                + "either because a value was already retrieved or "
                + "`freeze()` was directly called"
            )

    def _get(self, key: str) -> object:
        if key not in self.keys():
            raise KeyError(key)

        self._check_not_frozen()

        return getattr(self, key)

    def get(
        self, key: str, *, default: object = dataclasses.MISSING
    ) -> object:
        """Create and cache (if necessary) object with given key."""
        try:
            return self._get(key)
        except KeyError:
            if default is dataclasses.MISSING:
                raise
            else:
                return default

    def _check_before_set(self, key: str) -> None:
        self._check_not_frozen()

        if key not in self.keys():
            raise NewContainerKeyError(
                f"Cannot set new key on container: {key!r}"
            )

    def _set(self, key: str, value: object) -> None:
        self._check_before_set(key)

        setattr(self, key, value)

    def __getitem__(self, key: str) -> object:
        return nested_func(self, key, lambda ctr, key_part: ctr._get(key_part))

    def __contains__(self, key: str) -> bool:
        return nested_func(
            self, key, lambda ctr, key_part: key_part in ctr.keys()
        )

    def __setitem__(self, key: str, value: object) -> None:
        nested_func(self, key, lambda ctr, key_part: ctr._set(key_part, value))

    @override
    def __setattr__(self, key: str, value: object) -> None:
        if key in PRIVATE_CONTAINER_FIELD_NAMES:
            return super().__setattr__(key, value)

        self._check_before_set(key)

        return super().__setattr__(key, value)

    @override
    def __hash__(self) -> int:
        return hash(self.__class__)

    @classmethod
    def _create(
        cls: Type[TC],
        ctr_cache: Dict[Type[Container], Container],
        params: Dict[Type[Container], Dict[str, object]] | None = None,
    ) -> TC:
        try:
            return cast(TC, ctr_cache[cls])
        except KeyError:
            pass

        cls_params = params.get(cls) if params is not None else None
        cls_annotations = get_type_hints(cls)

        field_keys: Set[str] = set()
        child_ctrs: Dict[str, Container] = {}

        ctr_kwargs: Dict[str, object] = {}
        for field in dataclasses.fields(cls):
            if field.name in PRIVATE_CONTAINER_FIELD_NAMES:
                continue
            else:
                field_keys.add(field.name)

            field_annotation = cls_annotations[field.name]

            if isinstance(field_annotation, type) and issubclass(
                field_annotation, Container
            ):
                if (
                    field.default is not dataclasses.MISSING
                    or field.default_factory is not dataclasses.MISSING
                ):
                    raise ValueError(
                        "Cannot set defaults for child containers"
                    )

                child_ctr = field_annotation._create(ctr_cache, params=params)
                child_ctrs[field.name] = child_ctr
                ctr_kwargs[field.name] = child_ctr

        # NB: If there are either extra or missing params for this class,
        # Python will raise a `TypeError` when we construct it below.
        if cls_params is not None:
            ctr_kwargs.update(cls_params)

        cls._field_keys = field_keys
        cls._child_ctr_keys = set(child_ctrs)
        ctr = cls(**ctr_kwargs)

        for child_ctr in child_ctrs.values():
            child_ctr._parent_ctrs.add(ctr)

        ctr_cache[cls] = ctr
        return ctr

    @classmethod
    def create(
        cls: Type[TC],
        params: Dict[Type[Container], Dict[str, object]] | None = None,
    ) -> TC:
        """Create container and its child containers (cached by type).

        E.g.:

        ```python
        ctr = ParentContainer.create({ChildContainer: {"foo": 123}})
        ```
        """
        return cls._create(ctr_cache={}, params=params)


TC = TypeVar("TC", bound=Container)

if TYPE_CHECKING:
    ContainerWeakSet = weakref.WeakSet[Container]
else:
    ContainerWeakSet = weakref.WeakSet


@dataclasses.dataclass(frozen=True)
class PropertyValue(abc.ABC, Generic[TC, R]):
    func: Callable[[TC], R]

    @property
    def key(self) -> str:
        return self.func.__name__

    @abc.abstractmethod
    def _get(self, obj: TC) -> R: ...

    def __set_name__(self, owner: Type[TC], name: str) -> None:
        if not hasattr(owner, "_property_keys"):
            owner._property_keys = set()
        owner._property_keys.add(self.key)

    @overload
    def __get__(self, obj: None, obj_type: Type[TC]) -> Self: ...

    @overload
    def __get__(self, obj: TC, obj_type: Type[TC] | None = None) -> R: ...

    def __get__(
        self, obj: TC | None, obj_type: Type[TC] | None = None
    ) -> R | Self:
        if obj is None:
            return self

        return self._get(obj)

    def __set__(self, obj: TC, value: R) -> None:
        obj._check_before_set(self.key)

        obj._instance_cache[self.key] = value


@dataclasses.dataclass(frozen=True)
class Prototype(PropertyValue[TC, R]):
    @override
    def _get(self, obj: TC) -> R:
        obj.freeze()

        # Even though we never cache the result of this func call,
        # we use the instance cache to communicate that it's been
        # perturbed by the user.
        try:
            return cast(R, obj._instance_cache[self.key])
        except KeyError:
            pass

        return self.func(obj)


@dataclasses.dataclass(frozen=True)
class Singleton(PropertyValue[TC, R]):
    @override
    def _get(self, obj: TC) -> R:
        obj.freeze()

        try:
            value = cast(R, obj._instance_cache[self.key])
        except KeyError:
            value = self.func(obj)
            obj._instance_cache[self.key] = value

        return value


def call(func: Callable[[TC], R]) -> Prototype[TC, R]:
    """Call this method every time this object is retrieved."""
    return Prototype(func)


def cache(func: Callable[[TC], R]) -> Singleton[TC, R]:
    """Call this method once upon first retrieval and cache for later use."""
    return Singleton(func)


def container(cls: Type[T]) -> Type[T]:
    """Decorate container to enable field values."""
    return dataclasses.dataclass(frozen=False, unsafe_hash=True)(cls)



from dilib.experimental.core import Container as Container
from dilib.experimental.core import ContainerError as ContainerError
from dilib.experimental.core import (
    ContainerParamsError as ContainerParamsError,
)
from dilib.experimental.core import (
    FrozenContainerError as FrozenContainerError,
)
from dilib.experimental.core import (
    NewContainerKeyError as NewContainerKeyError,
)
from dilib.experimental.core import cache as cache
from dilib.experimental.core import call as call
from dilib.experimental.core import container as container



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
from __future__ import annotations

from typing import Any

import dilib.version

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "dilib"
copyright = "2025, dilib"  # noqa: A001
author = "dilib"

version = dilib.version.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = [
    "**/dilib.rst",
    "**/modules.rst",
]

language = "en"

# See https://www.sphinx-doc.org/en/master/usage/markdown.html
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# See https://myst-parser.readthedocs.io/en/latest/syntax/optional.html#auto-generated-header-anchors
myst_heading_anchors = 3

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = "dilib"
html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

if "dev" in version:
    version_match = "dev"
else:
    version_match = f"v{version}"

html_theme_options = {
    # Collapses navigation items
    "collapse_navigation": True,
    # Adjusts depth of sidebar links
    "navigation_depth": 2,
    # See https://github.com/pydata/pydata-sphinx-theme/blob/main/docs/conf.py
    "header_links_before_dropdown": 10,
    # Add the dropdown to the navbar
    "navbar_align": "left",
    "navbar_center": ["version-switcher", "navbar-nav"],
    # "show_version_warning_banner": True,
    "switcher": {
        # URL to versions file
        "json_url": "_static/versions.json",
        # Current version
        "version_match": version_match,
    },
    # See https://fontawesome.com/
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ansatzcapital/dilib",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/dilib",
            "icon": "fa-solid fa-cube",
        },
    ],
}

html_sidebars: dict[str, Any] = {
    "**": []  # No sidebar content for any page
}



# mypy: disable-error-code="comparison-overlap"
from __future__ import annotations

import dataclasses
import types
from typing import Any, TypeVar, cast

import pytest
from typing_extensions import assert_type

import dilib
import dilib.specs

TC = TypeVar("TC", bound=dilib.Config)


def get_config(
    config_cls: type[TC], more_type_safe: bool, **global_inputs: Any
) -> TC:
    if more_type_safe:
        return dilib.get_config(config_cls, **global_inputs)
    else:
        return cast(TC, config_cls().get(**global_inputs))


@dataclasses.dataclass(frozen=True)
class ValueWrapper:
    value: Any


@dataclasses.dataclass(frozen=True)
class ValuesWrapper:
    x: Any
    y: Any
    z: Any


@dataclasses.dataclass(frozen=True)
class SingletonValueWrapper(dilib.SingletonMixin, ValueWrapper):
    pass


@dataclasses.dataclass(frozen=True)
class PrototypeValueWrapper(dilib.PrototypeMixin, ValueWrapper):
    pass


with dilib.config_context():

    class BasicConfig(dilib.Config):
        x = dilib.Object(1)
        y: int = dilib.Prototype(lambda x, offset: x + offset, x, offset=1)

        foo = SingletonValueWrapper(value=x)
        bar = PrototypeValueWrapper(value=y)


def test_config_spec() -> None:
    # No inputs
    assert BasicConfig() == BasicConfig()
    assert hash(BasicConfig()) == hash(BasicConfig())

    # Basic inputs
    assert BasicConfig(x=1, y="hi") == BasicConfig(x=1, y="hi")
    assert BasicConfig(x=1, y="hi") != BasicConfig()

    assert hash(BasicConfig(x=1, y="hi")) == hash(BasicConfig(x=1, y="hi"))
    assert hash(BasicConfig(x=1, y="hi")) != hash(BasicConfig())


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_basic(more_type_safe: bool) -> None:
    config = get_config(BasicConfig, more_type_safe=more_type_safe)

    assert config._get_spec("x").obj == 1
    assert isinstance(config._get_spec("y").func_or_type, types.LambdaType)
    assert config._get_spec("foo").func_or_type is SingletonValueWrapper
    assert config._get_spec("bar").func_or_type is PrototypeValueWrapper


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_basic(more_type_safe: bool) -> None:
    config0: BasicConfig = get_config(
        BasicConfig, more_type_safe=more_type_safe
    )

    config0.x = 2
    spec_x = config0._get_spec("x")
    assert isinstance(spec_x, dilib.specs._Object)
    assert spec_x.obj == 2

    # Note that there are no class-level interactions, so if we
    # create a new instance, it doesn't have prior perturbations.
    config1 = dilib.get_config(BasicConfig)

    assert config1._get_spec("x").obj == 1


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_after_freeze(more_type_safe: bool) -> None:
    config = get_config(BasicConfig, more_type_safe=more_type_safe)

    config.freeze()
    with pytest.raises(dilib.FrozenConfigError):
        config.x = 100


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_add_key_after_load(more_type_safe: bool) -> None:
    config = get_config(BasicConfig, more_type_safe=more_type_safe)

    with pytest.raises(dilib.NewKeyConfigError):
        config.new_x = 100


with dilib.config_context():

    class ParentConfig0(dilib.Config):
        basic_config = BasicConfig()

        baz0 = SingletonValueWrapper(basic_config.x)

    class ParentConfig1(dilib.Config):
        basic_config = BasicConfig()

        baz1 = SingletonValueWrapper(basic_config.x)
        some_str1 = dilib.Object("abc")

    class GrandParentConfig(dilib.Config):
        parent_config0 = ParentConfig0()
        parent_config1 = ParentConfig1()

        foobar = SingletonValueWrapper(parent_config0.basic_config.x)
        some_str0 = dilib.Object("hi")

    class ErrorGrandParentConfig(dilib.Config):
        parent_config0 = ParentConfig0()
        parent_config1 = ParentConfig1()

        # This is pointing to a non-existent attr, so we will fail when
        # trying to get foobar via the container.
        foobar = dilib.Forward(parent_config0.non_existent_field)


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_dir(more_type_safe: bool) -> None:
    config = get_config(GrandParentConfig, more_type_safe=more_type_safe)

    assert dir(config) == [
        "foobar",
        "parent_config0",
        "parent_config1",
        "some_str0",
    ]
    assert dir(config.parent_config0) == ["basic_config", "baz0"]


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_nested_config(more_type_safe: bool) -> None:
    config = get_config(GrandParentConfig, more_type_safe=more_type_safe)

    assert id(config.parent_config0.basic_config) == id(
        config.parent_config1.basic_config
    )

    assert_type(config.parent_config0.basic_config, BasicConfig)
    assert_type(config.parent_config0.basic_config.bar, PrototypeValueWrapper)


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_config_attrs(more_type_safe: bool) -> None:
    config = get_config(GrandParentConfig, more_type_safe=more_type_safe)

    config.some_str0 = "hello"
    config.parent_config0.basic_config.x = 100
    config.parent_config1.some_str1 = "def"

    assert config._get_spec("some_str0").obj == "hello"
    assert config.parent_config1.basic_config._get_spec("x").obj == 100
    assert config.parent_config1._get_spec("some_str1").obj == "def"


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_config_strs(more_type_safe: bool) -> None:
    config = get_config(GrandParentConfig, more_type_safe=more_type_safe)

    config["some_str0"] = "hello"
    config["parent_config0.basic_config.x"] = 100
    config["parent_config1.some_str1"] = "def"

    assert config["some_str0"].obj == "hello"
    assert config["parent_config1.basic_config.x.obj"] == 100
    assert config["parent_config1.some_str1"].obj == "def"


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_perturb_nested_child_config(more_type_safe: bool) -> None:
    config = get_config(GrandParentConfig, more_type_safe=more_type_safe)

    with pytest.raises(dilib.SetChildConfigError):
        config.parent_config0 = cast(ParentConfig0, ParentConfig1())


class InputConfig0(dilib.Config):
    name = dilib.GlobalInput(str)
    context = dilib.GlobalInput(str, default="default")
    x = dilib.LocalInput(int)


class InputConfig1(dilib.Config):
    input_config0 = InputConfig0(x=1)

    y = dilib.Prototype(
        lambda x, offset: x + offset, input_config0.x, offset=1
    )


class BadInputConfig(dilib.Config):
    input_config0 = InputConfig0()  # Note missing inputs


@pytest.mark.parametrize("more_type_safe", [True, False])
def test_input_config(more_type_safe: bool) -> None:
    with pytest.raises(dilib.InputConfigError):
        InputConfig1().get()

    with pytest.raises(dilib.InputConfigError):
        InputConfig1().get(name=1)

    with pytest.raises(dilib.InputConfigError):
        BadInputConfig().get(name="hi")

    config = get_config(InputConfig1, name="hi", more_type_safe=more_type_safe)

    assert config.input_config0._get_spec("name").obj == "hi"
    assert config.input_config0._get_spec("context").obj == "default"
    assert config.input_config0._get_spec("x").obj == 1


class CollectionConfig(dilib.Config):
    x = dilib.Object(1)
    y = dilib.Object(2)
    z = dilib.Object(3)

    foo_tuple: tuple[int] = dilib.SingletonTuple(x, y)
    foo_list: list[int] = dilib.SingletonList(x, y)
    foo_dict_kwargs: dict[str, int] = dilib.SingletonDict(x=x, y=y)
    foo_dict_values0: dict[int, int] = dilib.SingletonDict({1: x, 2: y})
    foo_dict_values1: dict[str, int] = dilib.SingletonDict(values=x)
    foo_dict_values2: dict[str, int] = dilib.SingletonDict(
        {"x": x, "y": y}, z=z
    )

    # Check that untyped values don't trigger mypy errors
    _untyped_foo_tuple = dilib.SingletonTuple(x, y)
    _untyped_foo_list = dilib.SingletonList(x, y)
    _untyped_foo_dict_kwargs = dilib.SingletonDict(x=x, y=y)
    _untyped_foo_dict_values0: dict[int, int] = dilib.SingletonDict(
        {1: x, 2: y}
    )


class AnonymousConfig(dilib.Config):
    x = dilib.Singleton(ValueWrapper, 1)
    y = dilib.Singleton(ValueWrapper, dilib.Singleton(ValueWrapper, x))
    z = dilib.Singleton(ValueWrapper, dilib.Prototype(ValueWrapper, x))


class WrapperConfig(dilib.Config):
    _value = dilib.Singleton(ValueWrapper, 1)
    value = dilib.Singleton(ValueWrapper, _value)


class ForwardConfig(dilib.Config):
    other_config = GrandParentConfig()

    x = dilib.Forward(other_config.parent_config0.basic_config.x)
    x_value = dilib.Singleton(ValueWrapper, value=x)

    foo = dilib.Forward(other_config.parent_config0.basic_config.foo)
    foo_value = dilib.Singleton(ValueWrapper, value=foo)


class PartialKwargsConfig(dilib.Config):
    x = dilib.Object(1)
    y = dilib.Object(2)

    partial_kwargs = dilib.SingletonDict(x=x, y=y)

    values = dilib.Singleton(  # type: ignore[call-arg]
        ValuesWrapper,
        z=x,
        __lazy_kwargs=partial_kwargs,  # pyright: ignore
    )


class PartialKwargsOtherConfig(dilib.Config):
    partial_kwargs_config = PartialKwargsConfig()

    z = dilib.Object(3)
    values = dilib.Singleton(  # type: ignore[call-arg]
        ValuesWrapper,
        z=z,
        __lazy_kwargs=partial_kwargs_config.partial_kwargs,  # pyright: ignore
    )


def test_extra_global_inputs() -> None:
    with pytest.raises(dilib.InputConfigError):
        try:
            InputConfig1().get(name="testing", foobar=123)
        except dilib.InputConfigError as exc:
            assert "extra" in str(exc) and "'foobar'" in str(exc)
            raise


class InputConfigWithCollision(dilib.Config):
    input_config0 = InputConfig0(x=1)

    # NB: `name` collides with `input_config0.name`.
    name = dilib.GlobalInput(str)


def test_global_input_collisions() -> None:
    with pytest.raises(dilib.InputConfigError):
        try:
            InputConfigWithCollision().get(name="testing")
        except dilib.InputConfigError as exc:
            assert "collision" in str(exc) and "'name'" in str(exc)
            raise


def test_typing() -> None:
    # Would trigger mypy error:
    # cfg0: ParentConfig1 = dilib.get_config(ParentConfig0)

    cfg0: ParentConfig1 = dilib.get_config(ParentConfig1)

    # Would trigger mypy error:
    # _0: str = cfg0.basic_config.x

    _0: int = cfg0.basic_config.x  # noqa: F841
    _1: int = cfg0.basic_config.y  # noqa: F841



from __future__ import annotations

import datetime
from typing import Any, Final

import dilib.errors

PRIMITIVE_TYPES: Final[tuple[type, ...]] = (
    type(None),
    bool,
    int,
    float,
    str,
    datetime.date,
    datetime.time,
    datetime.datetime,
)


def check_type(
    value: Any, type_: type | None = None, desc: str | None = None
) -> None:
    """Check that value is of given type and raise error if not.

    Args:
        value: Value to check.
        type_: Type to check against.
        desc: Description for error.

    >>> import pytest; import dilib
    >>> check_type("abc", str)
    >>> with pytest.raises(dilib.InputConfigError):
    ...    check_type("abc", int)
    """
    if type_ is None:
        return

    if hasattr(type_, "__args__"):
        # TODO! Check nested typing types here.
        return
    else:
        types = (type_,)

    if not isinstance(value, types):
        raise dilib.errors.InputConfigError(
            f"{desc} input mismatch types: {type(value)} is not {type_}"
        )


def nested_getattr(obj: Any, address: str) -> Any:
    """Return last attr of obj specified by "."-separated address.

    >>> nested_getattr([], "__class__.__name__")
    'list'
    """
    for address_part in address.split("."):
        obj = getattr(obj, address_part)
    return obj


def nested_contains(obj: Any, address: str) -> bool:
    """Check existence of last attr of obj specified by "."-separated address.

    >>> class QuickAttrDict(dict):
    ...     def __getattr__(self, key):
    ...         result = self[key]
    ...         return (
    ...             QuickAttrDict(result) if isinstance(result, dict)
    ...             else result
    ...         )
    >>> values = QuickAttrDict({"b": {"c": {"d": 1}}})
    >>> nested_contains(values, "b.c.d")
    True
    >>> nested_contains(values, "b.c")
    True
    >>> nested_contains(values, "b")
    True
    >>> nested_contains(values, "b.c.e")
    False
    >>> nested_contains(values, "b.e")
    False
    >>> nested_contains(values, "e")
    False
    """
    split_address = address.split(".")
    last_address_part = split_address[-1]
    for address_part in split_address[:-1]:
        if address_part in obj:
            obj = getattr(obj, address_part)
        else:
            return False
    return last_address_part in obj


def nested_setattr(obj: Any, address: str, value: Any) -> None:
    """Set last attr of obj specified by "."-separated address to given value.

    >>> import unittest.mock
    >>> a = unittest.mock.MagicMock()
    >>> nested_setattr(a, "b.c", 123)
    >>> a.b.c
    123
    """
    split_address = address.split(".")
    for idx, address_part in enumerate(split_address):
        if idx < len(split_address) - 1:
            obj = getattr(obj, address_part)
        else:
            setattr(obj, address_part, value)



from __future__ import annotations

import abc
import dataclasses
import enum
from typing import Any

import pytest
from typing_extensions import override

from dilib.experimental import (
    Container,
    FrozenContainerError,
    NewContainerKeyError,
    cache,
    call,
    container,
)

#####################################################################
# Model Layer
#####################################################################


class Engine(abc.ABC):
    @abc.abstractmethod
    def start_engine(self) -> None: ...


@dataclasses.dataclass(frozen=True)
class MockEngine(Engine):
    @override
    def start_engine(self) -> None:
        print("Start mock engine")


@dataclasses.dataclass(frozen=True)
class DatabaseEngine(Engine):
    host: str
    port: int
    timeout_secs: int = 10

    @override
    def start_engine(self) -> None:
        print("Start db engine:", self.host, self.port, self.timeout_secs)


class TireType(enum.Enum):
    REGULAR = enum.auto()
    SPORT = enum.auto()
    SNOW = enum.auto()


@dataclasses.dataclass(frozen=True)
class Wheel:
    tire_type: TireType


class Car(abc.ABC):
    @abc.abstractmethod
    def start_car(self) -> None: ...


@dataclasses.dataclass(frozen=True)
class DefaultCar(Car):
    engine: Engine
    wheel0: Wheel
    wheel1: Wheel
    wheel2: Wheel
    wheel3: Wheel

    @override
    def start_car(self) -> None:
        self.engine.start_engine()


#####################################################################
# Container Layer
#####################################################################


@container
class CommonContainer(Container):
    env: str = "dev"


@container
class EngineContainer(Container):
    common_ctr: CommonContainer
    host: str
    port: int = 8000
    timeout_secs: int = 10

    @cache
    def engine(self) -> Engine:
        return DatabaseEngine(self.host, self.port, self.timeout_secs)


@container
class WheelContainer(Container):
    common_ctr: CommonContainer
    tire_type: TireType = TireType.REGULAR

    @call
    def wheel(self) -> Wheel:
        return Wheel(self.tire_type)


@container
class CarContainer(Container):
    common_ctr: CommonContainer
    engine_ctr: EngineContainer
    wheel_ctr: WheelContainer

    @cache
    def car(self) -> Car:
        return DefaultCar(
            engine=self.engine_ctr.engine,
            wheel0=self.wheel_ctr.wheel,
            wheel1=self.wheel_ctr.wheel,
            wheel2=self.wheel_ctr.wheel,
            wheel3=self.wheel_ctr.wheel,
        )


#####################################################################
# Application Layer
#####################################################################


def test_basic() -> None:
    # Create container with the minimal number of required params
    # (in this case, `EngineContainer` requires `host`).
    ctr = CarContainer.create({EngineContainer: {"host": "abc"}})

    # Child containers are cached by type across container hierarchy.
    assert ctr.common_ctr is ctr.engine_ctr.common_ctr
    assert ctr.common_ctr is ctr.wheel_ctr.common_ctr

    # Get objects and check types.
    engine = ctr.engine_ctr.engine
    assert isinstance(engine, DatabaseEngine)
    car = ctr.car
    assert isinstance(car, DefaultCar)

    assert ctr.car is ctr.car

    # Check that all the attributes are as expected.
    # Note that all values decorated with `@cache` (i.e., they're
    # singletons) are the same instance throughout, but all values
    # decorated with `@call` (i.e., they're prototypes)
    # get created on every call.
    assert engine is car.engine
    assert (
        car.wheel0.tire_type == TireType.REGULAR
        and car.wheel1.tire_type == TireType.REGULAR
        and car.wheel2.tire_type == TireType.REGULAR
        and car.wheel3.tire_type == TireType.REGULAR
    )
    assert car.wheel0 is not car.wheel1
    assert car.wheel0 is not car.wheel2
    assert car.wheel0 is not car.wheel3
    assert car.wheel0 is not ctr.wheel_ctr.wheel

    # We can't perturb the container once we've retrieved a value from it.
    with pytest.raises(FrozenContainerError):
        ctr.wheel_ctr.tire_type = TireType.SPORT


def test_typing() -> None:
    # We explicitly set types to show that the type checker understands
    # everything.
    ctr: CarContainer = CarContainer.create({EngineContainer: {"host": "abc"}})

    _0: CommonContainer = ctr.common_ctr
    _1: EngineContainer = ctr.engine_ctr
    _2: WheelContainer = ctr.wheel_ctr

    _3: Car = ctr.car
    _4: Engine = ctr.engine_ctr.engine
    _5: Wheel = ctr.wheel_ctr.wheel

    _6: str = ctr.engine_ctr.host
    _7: int = ctr.engine_ctr.port
    _8: int = ctr.engine_ctr.timeout_secs


def test_dict_like() -> None:
    ctr = CarContainer.create({EngineContainer: {"host": "abc"}})

    # Get field and property values via dict-like syntax, but with dotted keys.
    assert "engine_ctr.host" in ctr
    host = ctr["engine_ctr.host"]
    assert host == "abc"

    assert "engine_ctr.engine" in ctr
    engine = ctr["engine_ctr.engine"]
    assert isinstance(engine, DatabaseEngine)

    # Check what happens with non-existent keys.
    assert "engine_ctr.foo" not in ctr
    with pytest.raises(KeyError):
        ctr["foo"]
    with pytest.raises(KeyError):
        ctr["engine_ctr.foo"]

    # Check that we can't perturb via dotted keys either.
    with pytest.raises(FrozenContainerError):
        ctr["engine_ctr.engine"] = MockEngine()

    # Check keys are the top-level values.
    assert ctr.keys() == {
        # Child containers.
        "common_ctr",
        "engine_ctr",
        "wheel_ctr",
        # Property values.
        "car",
    }
    assert ctr.engine_ctr.keys() == {
        # Child containers.
        "common_ctr",
        # Field values
        "host",
        "port",
        "timeout_secs",
        # Property values.
        "engine",
    }


def test_ctr_params() -> None:
    # Check that our container params made their way through as expected.
    ctr = CarContainer.create(
        {
            EngineContainer: {"host": "abc"},
            WheelContainer: {"tire_type": TireType.SNOW},
        }
    )

    engine = ctr.engine_ctr.engine
    assert isinstance(engine, DatabaseEngine)
    car = ctr.car
    assert isinstance(car, DefaultCar)

    assert engine.host == "abc"
    assert car.wheel0.tire_type == TireType.SNOW

    # We should raise an error because we're missing the required
    # `EngineContainer` `host` param.
    with pytest.raises(TypeError):
        ctr = CarContainer.create(
            {WheelContainer: {"tire_type": TireType.SNOW}}
        )

    # We should raise an error because we have a typo and thought the
    # param was `tyre_type` instead of the correct `tire_type`.
    with pytest.raises(TypeError):
        ctr = CarContainer.create(
            {
                EngineContainer: {"host": "abc"},
                WheelContainer: {"tyre_type": TireType.SNOW},
            }
        )


def test_perturb_basic() -> None:
    # Perturb container after creation.
    ctr = CarContainer.create(
        {
            EngineContainer: {"host": "abc"},
            WheelContainer: {"tire_type": TireType.SNOW},
        }
    )

    # Subtle point: we can actually get field values before
    # perturbing because we know they can't depend on any other values.
    assert ctr.engine_ctr.host == "abc"

    # Pertub both field and property values.
    ctr["wheel_ctr.tire_type"] = TireType.SPORT
    ctr.engine_ctr.host = "def"
    ctr.engine_ctr.engine = MockEngine()

    car = ctr.car
    assert isinstance(car, DefaultCar)

    # Check that the perturbations override all other values.
    assert (
        car.wheel0.tire_type == TireType.SPORT
        and car.wheel1.tire_type == TireType.SPORT
        and car.wheel2.tire_type == TireType.SPORT
        and car.wheel3.tire_type == TireType.SPORT
    )
    assert ctr.engine_ctr.host == "def"
    assert isinstance(car.engine, MockEngine)
    assert car.engine is ctr.engine_ctr.engine

    # Check that we're frozen now.
    with pytest.raises(FrozenContainerError):
        ctr.engine_ctr.host = "xyz"

    # Every container is its own instance, i.e.,
    # there are no class-level interactions.
    car1 = CarContainer.create(
        {
            EngineContainer: {"host": "abc"},
            WheelContainer: {"tire_type": TireType.SNOW},
        }
    ).car
    assert car is not car1
    assert isinstance(car1, DefaultCar)
    assert car1.wheel0.tire_type == TireType.SNOW


def test_get_bad_key() -> None:
    ctr = CarContainer.create({EngineContainer: {"host": "abc"}})

    with pytest.raises(AttributeError):
        _ = ctr.foo  # type: ignore[attr-defined]

    with pytest.raises(KeyError):
        ctr["foo"]


def test_perturb_new_key() -> None:
    ctr = CarContainer.create({EngineContainer: {"host": "abc"}})

    with pytest.raises(NewContainerKeyError):
        ctr.foo = "abc"

    with pytest.raises(NewContainerKeyError):
        ctr["foo"] = "abc"


@dataclasses.dataclass(frozen=True)
class Foo:
    x: int
    y: float
    z: str


@dataclasses.dataclass(frozen=True)
class Bar:
    foo0: Foo
    foo1: Foo


@container
class ClassicMigrationContainer(Container):
    x: int = 1

    @cache
    def _foo_kwargs0(self) -> dict[str, Any]:
        # Example of delayed collection and partial/lazy kwargs.
        return {"x": self.x}

    @cache
    def _foo_kwargs1(self) -> dict[str, Any]:
        # Example of delayed collection and partial/lazy kwargs.
        return {"y": 10.0}

    @cache
    def foo0(self) -> Foo:
        # Example of combining partial/lazy kwargs.
        return Foo(**self._foo_kwargs0, **self._foo_kwargs1, z="abc")

    @cache
    def bar(self) -> Bar:
        # Second param here is an example of anonymous construction.
        return Bar(foo0=self.foo0, foo1=Foo(5, 20.0, "def"))

    @cache
    def forward_bar(self) -> Bar:
        # Example of forwarding.
        return self.bar

    @cache
    def bar_foo0(self) -> Foo:
        # Example of obj attr.
        return self.bar.foo0


def test_classic_migration() -> None:
    ctr = ClassicMigrationContainer.create()

    # This tests perturbation through partial kwargs and forwards.
    ctr._foo_kwargs0 = {"x": 2}

    assert ctr._foo_kwargs0 == {"x": 2}
    assert ctr.foo0 == Foo(2, 10.0, "abc")

    assert ctr.bar.foo0 is ctr.foo0
    assert ctr.bar.foo1 == Foo(5, 20.0, "def")

    assert ctr.bar is ctr.forward_bar
    assert ctr.bar_foo0 is ctr.foo0


@container
class VeryNestedContainer(Container):
    engine_ctr: EngineContainer

    @cache
    def foo(self) -> Foo:
        return Foo(1, 10.0, z=self.engine_ctr.common_ctr.env)


def test_perturb_very_nested_prototype() -> None:
    ctr = VeryNestedContainer.create({EngineContainer: {"host": "abc"}})

    ctr.engine_ctr.common_ctr.env = "prod"

    assert ctr.engine_ctr.common_ctr.env == "prod"
    assert ctr.foo.z == "prod"



from __future__ import annotations

import abc
import dataclasses
from typing import Any

from typing_extensions import override

import dilib


def get_db_value(_0: Any, _1: Any) -> bool:
    return False


class Seat(dilib.SingletonMixin):
    pass


class Engine(abc.ABC):
    @property
    @abc.abstractmethod
    def started(self) -> bool: ...

    @abc.abstractmethod
    def start(self) -> None: ...


@dataclasses.dataclass(frozen=True)
class DBEngine(Engine, dilib.SingletonMixin):
    db_address: str

    @property
    @override
    def started(self) -> bool:
        return get_db_value(self.db_address, "engine")

    @override
    def start(self) -> None:
        pass


class MockEngine(Engine, dilib.SingletonMixin):
    @property
    @override
    def started(self) -> bool:
        return True

    @override
    def start(self) -> None:
        pass


class Car(dilib.SingletonMixin):
    def __init__(self, seats: list[Seat], engine: Engine) -> None:
        self.seats = seats
        self.engine = engine

        self.state = 0

    def drive(self) -> None:
        if not self.engine.started:
            self.engine.start()
        self.state = 1

    def stop(self) -> None:
        self.state = 0


with dilib.config_context():

    class EngineConfig(dilib.Config):
        db_address = dilib.GlobalInput(type_=str, default="some-db-address")
        engine: Engine = DBEngine(db_address)

    class CarConfig(dilib.Config):
        engine_config = EngineConfig()

        seat0 = Seat()
        seat1 = Seat()
        seats = dilib.SingletonList(seat0, seat1)

        car = Car(seats, engine=engine_config.engine)


def test_basic_demo() -> None:
    config = dilib.get_config(CarConfig, db_address="some-db-address")
    container = dilib.get_container(config)

    car: Car = container.config.car
    assert isinstance(car, Car)
    assert id(car) == id(container.config.car)  # Because it's a Singleton
    assert isinstance(car.engine, DBEngine)


def test_perturb_demo() -> None:
    config = dilib.get_config(CarConfig, db_address="some-db-address")
    config.engine_config.engine = MockEngine()
    container = dilib.get_container(config)

    assert isinstance(container.config.car.engine, MockEngine)



"""Specs are recipes of how objects should be created.

The public functions are named like classes (e.g., :func:`Object`,
:func:`Singleton`), while the actual specs are private classes
with underscore prefixes. This mimics the pattern seen in
`dataclasses.field()` func vs. `dataclass.Field` class.

Note that we "trick" IDEs and static type checkers by casting
the spec objects into whatever type they will return when instantiated
by the container. Although a bit inelegant, this actually makes
all the config wiring type check exactly as expected.
"""

from __future__ import annotations

import contextlib
from typing import Any, Callable, Generator, Generic, TypeVar, cast

from typing_extensions import ParamSpec, Self, TypeAlias, override

import dilib.errors

MATERIALIZE = True

MISSING = object()

SpecID: TypeAlias = int
P = ParamSpec("P")
T = TypeVar("T")


def instantiate(cls: type[T], *args: Any, **kwargs: Any) -> T:
    """Instantiate obj from Spec parts.

    :meta private:
    """
    try:
        if issubclass(
            cls,
            (PrototypeMixin, SingletonMixin),
        ):
            obj = cls.__new__(cls, _materialize=True)
            obj.__init__(*args, **kwargs)  # noqa
            return cast(T, obj)

        return cls(*args, **kwargs)
    except TypeError as exc:
        raise TypeError(f"{cls}: {str(exc)}") from None


class AttrFuture:
    """Future representing attr access on a Spec by its spec id.

    :meta private:
    """

    def __init__(self, root_spec_id: SpecID, attrs: list[str]) -> None:
        self.root_spec_id = root_spec_id
        self.attrs = attrs

    def __getattr__(self, attr: str) -> AttrFuture:
        return AttrFuture(self.root_spec_id, self.attrs + [attr])


class Spec(Generic[T]):
    """Represents delayed object to be instantiated later.

    Use one of child classes when describing objects.
    """

    _INTERNAL_FIELDS = ["spec_id"]
    NEXT_SPEC_ID = 0

    def __init__(self, spec_id: SpecID | None = None) -> None:
        self.spec_id = self._get_next_spec_id() if spec_id is None else spec_id

    def __getattr__(self, attr: str) -> AttrFuture:
        return AttrFuture(self.spec_id, [attr])

    @override
    def __setattr__(self, name: str, value: Any) -> None:
        if (
            name.startswith("__")
            or name == "_INTERNAL_FIELDS"
            or name in self._INTERNAL_FIELDS
        ):
            return super().__setattr__(name, value)

        # NB: We considered supporting this kind of perturbation,
        # but the issue is that we don't know whether the config
        # this spec is attached to has been frozen. For sake of safety
        # and simplicity, we raise an error here instead.
        raise dilib.errors.PerturbSpecError(
            "Cannot set on a spec. "
            "If you'd like to perturb a value used by a spec, "
            "promote it to be a config field and perturb the config instead."
        )

    # For mypy
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return None

    @classmethod
    def _get_next_spec_id(cls) -> SpecID:
        # NB: Need to use `Spec` explicitly to ensure all `Spec`
        # subclasses share the same spec id space.
        result = Spec.NEXT_SPEC_ID
        Spec.NEXT_SPEC_ID += 1
        return result


class _Object(Spec[T]):
    """Represents fully-instantiated object to pass through."""

    _INTERNAL_FIELDS = Spec._INTERNAL_FIELDS + ["obj"]

    def __init__(self, obj: T, spec_id: SpecID | None = None) -> None:
        super().__init__(spec_id=spec_id)
        self.obj = obj


# noinspection PyPep8Naming
def Object(obj: T) -> T:  # noqa: N802
    """Spec to pass through a fully-instantiated object.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.Object(1)

    Args:
        obj: Fully-instantiated object to pass through.
    """
    # Cast because the return type will act like a T
    return cast(T, _Object(obj))


class _Input(Spec[T]):
    """Represents user input to config."""

    _INTERNAL_FIELDS = Spec._INTERNAL_FIELDS + ["type_", "default"]

    def __init__(
        self, type_: type[T] | None = None, default: Any = MISSING
    ) -> None:
        super().__init__()
        self.type_ = type_
        self.default = default


class _GlobalInput(_Input[T]):
    """Represents input passed in at config instantiation."""

    pass


# noinspection PyPep8Naming
def GlobalInput(  # noqa: N802
    type_: type[T] | None = None, default: Any = MISSING
) -> T:
    """Spec to use user input passed in at config instantiation.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.GlobalInput(type_=str)
    ...     y = dilib.GlobalInput(type_=int, default=1)

    Args:
        type_: Expected type of input, for both static and runtime check.
        default: Default value if no input is provided.
    """
    # Cast because the return type will act like a T
    return cast(T, _GlobalInput(type_=type_, default=default))


class _LocalInput(_Input[T]):
    """Represents input passed in at config declaration."""

    pass


# noinspection PyPep8Naming
def LocalInput(  # noqa: N802
    type_: type[T] | None = None, default: Any = MISSING
) -> T:
    """Spec to use user input passed in at config declaration.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.LocalInput(type_=str)
    ...     y = dilib.LocalInput(type_=int, default=1)

    >>> class BarConfig(dilib.Config):
    ...     foo_config = FooConfig(x="abc", y=123)

    >>> config = dilib.get_config(BarConfig)
    >>> container = dilib.get_container(config)

    >>> container.config.foo_config.x
    'abc'

    >>> container.config.foo_config.y
    123

    Args:
        type_: Expected type of input, for both static and runtime check.
        default: Default value if no input is provided.
    """
    # Cast because the return type will act like a `T`.
    return cast(T, _LocalInput(type_=type_, default=default))


class _Callable(Spec[T]):
    """Represents callable (e.g., func, type) to be called with given args."""

    _INTERNAL_FIELDS = Spec._INTERNAL_FIELDS + [
        "func_or_type",
        "args",
        "lazy_kwargs",
        "kwargs",
    ]

    def __init__(
        self,
        func_or_type: Callable[..., T],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.func_or_type = func_or_type
        self.args = args
        self.lazy_kwargs = kwargs.pop("__lazy_kwargs", None)
        self.kwargs = kwargs

    def instantiate(self) -> Any:
        """Instantiate spec into object."""
        if isinstance(self.func_or_type, type):
            # noinspection PyTypeChecker
            return instantiate(self.func_or_type, *self.args, **self.kwargs)
        else:
            # Non-type callable (e.g., function, functor).
            return self.func_or_type(*self.args, **self.kwargs)

    def copy_with(self, *args: Any, **kwargs: Any) -> Self:
        """Make a copy with replaced args.

        Used to replace arg specs with materialized args.
        """
        return self.__class__(self.func_or_type, *args, **kwargs)


class _Prototype(_Callable[T]):
    pass


# noinspection PyPep8Naming
def Prototype(  # noqa: N802
    func_or_type: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    """Spec to call with args and no caching.

    Can be used with anything callable, including types and functions.

    >>> class Foo:
    ...     def __init__(self, x: int) -> None:
    ...         self.x = x

    >>> class FooConfig(dilib.Config):
    ...     foo = dilib.Prototype(Foo, x=1)

    >>> config = dilib.get_config(FooConfig)
    >>> container = dilib.get_container(config)
    >>> assert container.config.foo is not container.config.foo
    """
    # Cast because the return type will act like a `T`.
    return cast(T, _Prototype(func_or_type, *args, **kwargs))


def _identity(obj: T) -> T:
    return obj


def _union_dict_and_kwargs(
    values: dict[str, Any], **kwargs: Any
) -> dict[str, Any]:
    new_values = values.copy()
    new_values.update(**kwargs)
    return new_values


# noinspection PyPep8Naming
def Forward(obj: T) -> T:  # noqa: N802
    """Spec to simply forward to other spec.

    Often useful as a switch to dispatch to other specs.

    >>> class FooConfig(dilib.Config):
    ...     x0 = dilib.Object(1)
    ...     x1 = dilib.Object(2)
    ...     x = dilib.Forward(x0)

    >>> config = dilib.get_config(FooConfig)
    >>> container = dilib.get_container(config)
    >>> container.config.x
    1

    The switch can be perturbed like:

    >>> config = dilib.get_config(FooConfig)
    >>> config.x = dilib.Forward(config.x1)
    >>> container = dilib.get_container(config)
    >>> container.config.x
    2
    """
    # Cast because the return type will act like a `T`.
    return cast(T, _Prototype(_identity, obj))


class _Singleton(_Callable[T]):
    pass


# noinspection PyPep8Naming
def Singleton(  # noqa: N802
    func_or_type: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    """Spec to call with args and caching per config field.

    Can be used with anything callable, including types and functions.

    >>> class Foo:
    ...     def __init__(self, x: int) -> None:
    ...         self.x = x

    >>> class FooConfig(dilib.Config):
    ...     foo = dilib.Singleton(Foo, x=1)

    >>> config = dilib.get_config(FooConfig)
    >>> container = dilib.get_container(config)
    >>> assert container.config.foo is container.config.foo
    """
    # Cast because the return type will act like a `T`.
    return cast(T, _Singleton(func_or_type, *args, **kwargs))


# noinspection PyPep8Naming
def SingletonTuple(*args: T) -> tuple[T]:  # noqa: N802
    """Spec to create tuple with args and caching per config field.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.Object(1)
    ...     y = dilib.Object(2)
    ...     values = dilib.SingletonTuple(x, y)
    """
    # Cast because the return type will act like a tuple of `T`.
    return cast("tuple[T]", _Singleton(tuple, args))


# noinspection PyPep8Naming
def SingletonList(*args: T) -> list[T]:  # noqa: N802
    """Spec to create list with args and caching per config field.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.Object(1)
    ...     y = dilib.Object(2)
    ...     values = dilib.SingletonList(x, y)
    """
    # Cast because the return type will act like a list of `T`.
    return cast("list[T]", _Singleton(list, args))


# noinspection PyPep8Naming
def SingletonDict(  # noqa: N802
    values: dict[Any, T] | None = None, /, **kwargs: T
) -> dict[Any, T]:
    """Spec to create dict with args and caching per config field.

    Can specify either by pointing to a dict, passing in kwargs,
    or unioning both.

    >>> class FooConfig(dilib.Config):
    ...     x = dilib.Object(1)
    ...     y = dilib.Object(2)
    ...     values = dilib.SingletonDict(x=x, y=y)
    ...     # Equivalent to:
    ...     also_values = dilib.SingletonDict({"x": x, "y": y})
    """
    if values is None:
        # Cast because the return type will act like a dict of `T`.
        return cast("dict[Any, T]", _Singleton(dict, **kwargs))
    else:
        # Cast because the return type will act like a dict of `T`.
        return cast(
            "dict[Any, T]",
            _Singleton(_union_dict_and_kwargs, values, **kwargs),
        )


@contextlib.contextmanager
def config_context() -> Generator[None, None, None]:
    """Enable delayed mode for `PrototypeMixin` and `SingletonMixin`.

    >>> class Foo(dilib.SingletonMixin):
    ...     def __init__(self, x: int) -> None:
    ...         self.x = x

    >>> with dilib.config_context():
    ...     class FooConfig(dilib.Config):
    ...         foo = Foo(x=1)
    """
    global MATERIALIZE
    MATERIALIZE = False
    yield
    MATERIALIZE = True


class PrototypeMixin:
    """Helper class for `Prototype` to ease syntax in `Config`.

    Equivalent to `dilib.Prototype(cls, ...)`.

    See:
        * :class:`Prototype`
        * :func:`config_context`
    """

    def __new__(cls: type[Any], *args: Any, **kwargs: Any) -> Any:
        if MATERIALIZE:
            # noinspection PyTypeChecker
            return super().__new__(cls)
        else:
            return Prototype(cls, *args, **kwargs)


class SingletonMixin:
    """Helper class for `Singleton` to ease syntax in `Config`.

    Equivalent to `dilib.Singleton(cls, ...)`.

    See:
        * :class:`Singleton`
        * :func:`config_context`
    """

    def __new__(cls: type[Any], *args: Any, **kwargs: Any) -> Any:
        if MATERIALIZE:
            # noinspection PyTypeChecker
            return super().__new__(cls)
        else:
            return Singleton(cls, *args, **kwargs)



from __future__ import annotations

import abc
from typing import Sequence, TypeVar

from typing_extensions import override

import dilib

T = TypeVar("T")


class BaseMultiplier(abc.ABC):
    @abc.abstractmethod
    def get_result(self) -> int: ...


class SimpleMultiplier(BaseMultiplier):
    def __init__(self, x: int, y: int = 2) -> None:
        self.x = x
        self.y = y

    @override
    def get_result(self) -> int:
        return self.x * self.y


class MockMultipler(BaseMultiplier):
    @override
    def get_result(self) -> int:
        return 42


def add_ints(x: int, y: int) -> int:
    return x + y


def test_typing() -> None:
    spec0: int = dilib.Object(1)
    spec1: int = dilib.Object(2)
    spec2: str = dilib.Object("abc")

    _3: int = dilib.GlobalInput(type_=int)
    _4: str = dilib.LocalInput(type_=str)
    _5: Sequence[str] = dilib.LocalInput(type_=Sequence[str])

    _6: BaseMultiplier = dilib.Prototype(MockMultipler)
    _7: BaseMultiplier = dilib.Prototype(SimpleMultiplier, 100)
    _8: BaseMultiplier = dilib.Singleton(SimpleMultiplier, 100, y=3)
    _9: SimpleMultiplier = dilib.Singleton(SimpleMultiplier, 100, y=3)

    _11: Sequence[int] = dilib.SingletonList(spec0, spec0)
    _12: tuple[str, ...] = dilib.SingletonTuple(spec2, spec2)
    # TODO: Support more narrow `tuple` types?
    # _13: tuple[str, str] = dilib.SingletonTuple(spec2, spec2)
    _14: dict[str, int] = dilib.SingletonDict(a=spec0, b=spec1)

    # Would cause mypy error:
    # _15: str = dilib.Singleton(add_ints, 1, "abc")
    _15: int = dilib.Singleton(add_ints, 1, 2)
